﻿using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.MoveVectorEncoding.Interfaces;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Data;
using HeuristicLab.Data.MoveVectorData;

namespace HeuristicLab.Encodings.MoveVectorEncoding.Manipulators
{
    [Item("Multi Move Manipulator", "Randomly selects and applies one of its manipulators every time it is called.")]
    [StorableClass]
    public class MultiMoveManipulator : StochasticMultiBranch<IMoveVectorManipulator>, IMoveVectorManipulator, IBoundedMoveVectorOperator, IStochasticOperator
    {
        public override bool CanChangeName
        {
            get { return false; }
        }
        protected override bool CreateChildOperation
        {
            get { return true; }
        }

        public ILookupParameter<MoveVector> MoveVectorParameter
        {
            get { return (ILookupParameter<MoveVector>)Parameters["MoveVector"]; }
        }

        public IValueLookupParameter<IntMatrix> BoundsParameter
        {
            get { return (IValueLookupParameter<IntMatrix>)Parameters["Bounds"]; }
        }

        [StorableConstructor]
        protected MultiMoveManipulator(bool deserializing) : base(deserializing) { }
        protected MultiMoveManipulator(MultiMoveManipulator original, Cloner cloner) : base(original, cloner) { }
        public MultiMoveManipulator()
          : base()
        {
            Parameters.Add(new LookupParameter<MoveVector>("MoveVector", "The encoding vector that is to be manipulated."));
            Parameters.Add(new ValueLookupParameter<IntMatrix>("Bounds", "The bounds matrix can contain one row for each dimension with three columns specifying minimum (inclusive), maximum (exclusive), and step size. If less rows are given the matrix is cycled."));
            foreach (Type type in ApplicationManager.Manager.GetTypes(typeof(IMoveVectorManipulator), typeof(MoveVectorEncoding).Assembly))
            {
                if (!typeof(MultiOperator<IMoveVectorManipulator>).IsAssignableFrom(type))
                    Operators.Add((IMoveVectorManipulator)Activator.CreateInstance(type), true);
            }
            SelectedOperatorParameter.ActualName = "SelectedManipulationOperator";
        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new MultiMoveManipulator(this, cloner);
        }

        protected override void Operators_ItemsReplaced(object sender, CollectionItemsChangedEventArgs<IndexedItem<IMoveVectorManipulator>> e)
        {
            base.Operators_ItemsReplaced(sender, e);
            ParameterizeManipulators(e.Items);
        }

        protected override void Operators_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<IMoveVectorManipulator>> e)
        {
            base.Operators_ItemsAdded(sender, e);
            ParameterizeManipulators(e.Items);
        }

        private void ParameterizeManipulators(IEnumerable<IndexedItem<IMoveVectorManipulator>> manipulators)
        {
            foreach (var m in manipulators.Select(x => x.Value))
            {
                m.MoveVectorParameter.ActualName = MoveVectorParameter.Name;
                var stOp = m as IStochasticOperator;
                if (stOp != null) stOp.RandomParameter.ActualName = RandomParameter.Name;

                var bounded = m as IBoundedMoveVectorOperator;
                if (bounded != null) bounded.BoundsParameter.ActualName = MoveVectorParameter.Name + "." + BoundsParameter.Name;
            }
        }

        public override IOperation InstrumentedApply()
        {
            if (Operators.Count == 0) throw new InvalidOperationException(Name + ": Please add at least one move manipulator to choose from.");
            return base.InstrumentedApply();
        }
    }
}